import pathlib

import numpy as np

import radontea

import sinogram


def test_2d_backproject():
    sino, angles = sinogram.create_test_sino(A=100, N=100)
    r = radontea.backproject(sino, angles, padding=True)

    # np.savetxt('outfile.txt', np.array(r).flatten().view(float), fmt="%.8f")
    reffile = pathlib.Path(__file__).parent / "data" / "2d_backproject.txt"
    ref = np.loadtxt(str(reffile))
    assert np.allclose(np.array(r).flatten().view(float), ref)


def test_3d_backproject():
    A = 100
    M = 30
    N = 100

    sino, angles = sinogram.create_test_sino(A=100, N=100)
    sino3d = np.zeros((A, M, N), dtype=float)
    for ii in range(M):
        sino3d[:, ii, :] = sino
    r = radontea.backproject_3d(sino3d, angles, padding=True)
    reffile = pathlib.Path(__file__).parent / "data" / "2d_backproject.txt"
    ref = np.loadtxt(str(reffile))
    assert np.allclose(np.array(r[:, 0, :]).flatten().view(float), ref)
    assert np.all(r[:, 0, :] == r[:, 1, :])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# project documentation build configuration file, created by
# sphinx-quickstart on Sat Feb 22 09:35:49 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os.path as op
import sys

import radontea

# include parent directory
pdir = op.dirname(op.dirname(op.abspath(__file__)))
sys.path.insert(0, pdir)
# include extenstions
sys.path.append(op.abspath('extensions'))


# http://www.sphinx-doc.org/en/stable/ext/autodoc.html#confval-autodoc_member_order
# Order class attributes and functions in separate blocks
autodoc_member_order = 'bysource'

# Display link to GitHub repo instead of doc on rtfd
rst_prolog = """
:github_url: https://github.com/RI-imaging/radontea
"""

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.


extensions = ['sphinx.ext.intersphinx',
              'sphinx.ext.autosummary',
              'sphinx.ext.autodoc',
              'sphinx.ext.mathjax',
              'sphinx.ext.napoleon',
              'fancy_include',
              'github_changelog',
              ]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'radontea'
copyright = '2017, Paul Müller'
author = 'Paul Müller'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# The full version, including alpha/beta/rc tags.
version = radontea.__version__
release = version


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'default'

# Output file base name for HTML help builder.
htmlhelp_basename = project+'doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', project+'.tex', project+' Documentation',
     author, 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', project, project+' Documentation',
     author, 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', project, project + ' Documentation',
     author, project,
     "Methods for the inversion of the Radon transform",
     'Numeric'),
]

# -----------------------------------------------------------------------------
# intersphinx
# -----------------------------------------------------------------------------
intersphinx_mapping = {
    "python": ('https://docs.python.org/', None),
    "numpy": ('http://docs.scipy.org/doc/numpy', None),
    "scipy": ('https://docs.scipy.org/doc/scipy/reference/', None),
    }



import numpy as np


def integrate(sinogram: np.ndarray, angles: np.ndarray, count=None,
              max_count=None) -> np.ndarray:
    """2D sum-reconstruction with the Fourier slice theorem

    Computes the inverse of the Radon transform by computing the
    integral in real space.


    Parameters
    ----------
    sinogram: (A,N) ndarray
        Two-dimensional sinogram of line recordings.
    angles: (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.


    Returns
    -------
    out: ndarray
        The reconstructed image.
    """
    if max_count is not None:
        with max_count.get_lock():
            max_count.value += sinogram.shape[1]**2 + 1
    # In the script we used the unitary angular frequency (uaf) Fourier
    # Transform. The discrete Fourier transform is equivalent to the
    # unitary ordinary frequency (uof) Fourier transform.
    #
    # uof: f₁(ξ) = int f(x) exp(-2πi xξ)
    #
    # uaf: f₃(ω) = (2π)^(-n/2) int f(x) exp(-i ωx)
    #
    # f₁(ξ) = (2π)^(n/2) f₃(ω)
    # ω = 2πξ
    #
    # We have a one-dimensional (n=1) Fourier transform and UB in the
    # script is equivalent to f₃(ω). Because we are working with the
    # uaf, we divide by sqrt(2π) after computing the fft with the uof.

    # Corresponding sample frequencies
    fx = np.fft.fftfreq(sinogram[0].shape[0])  # 1D array
    kx = 2 * np.pi * fx

    # Get the angles ϕ₀.
    phi0 = (angles).reshape(-1, 1)
    # Differentials for integral
    dphi0 = len(angles) / np.pi
    dkx = np.abs(kx[1] - kx[0])

    # We will later multiply with phi0.
    # Make sure we are using correct shapes
    kx = kx.reshape(1, -1)

    # Create the integrand
    # Integrals over ϕ₀ [0,2π]; kx [-k,k]
    #   - We do not have double coverage, as the input data is only
    #     from 0 to 180 degrees.
    #   - unitary angular frequency to unitary ordinary frequency
    #     conversion performed in calculation of P=FT(sinogram).
    #
    # f(r) = 1 / ( (2π)^(3/2) )                     (prefactor)
    #      * iint dϕ₀ dkx                           (prefactor)
    #      * |kx|                                   (prefactor)
    #      * P_ϕ₀(kx)                               (dependent on ϕ₀)
    #      * exp( i kx (cos(ϕ₀)*x + sin(ϕ₀)*y ) )   (dependent on ϕ₀ and r)
    #
    # everything that is not dependent on phi0:
    prefactor = 1 / ((2 * np.pi)**(3 / 2))
    prefactor *= dphi0 * dkx
    prefactor *= np.abs(kx)

    # Initiate function f
    N = len(sinogram[0])
    coords = np.linspace(-N / 2., N / 2., N, endpoint=False)
    x, y = np.meshgrid(coords, coords)
    points = np.zeros((N**2, 2))
    points[:, 0] = x.flatten()
    points[:, 1] = y.flatten()
    f = np.zeros(len(points), dtype=np.complex128)
    lenf = len(f)

    # Initiate vector r that corresponds to calculating a value of f.
    r = np.zeros((2, 1, 1), dtype=np.complex128)

    # Compute the Fourier transform of uB.
    # This is true: np.fft.fft(UB)[0] == np.fft.fft(UB[0])
    # because axis -1 is always used.
    P = np.fft.fft(np.fft.ifftshift(sinogram, axes=-1)) / np.sqrt(2 * np.pi)

    if count is not None:
        with count.get_lock():
            count.value += 1

    for j in range(lenf):
        # Get r (We compute f(r) in this for-loop)
        r[0][:] = points[j][0]  # x
        r[1][:] = points[j][1]  # y

        # Integrand changes with r, so we have to create a new
        # array:
        integrand = prefactor * P

        # Reminder:
        # f(r) = 1 / ( (2π)^(3/2) )                    (prefactor)
        #      * iint dϕ₀ dkx                          (prefactor)
        #      * |kx|                                  (prefactor)
        #      * P_ϕ₀(kx)                              (dependent on ϕ₀)
        #      * exp( i kx (cos(ϕ₀)*x + sin(ϕ₀)*y ) )  (dependent on ϕ₀ and r)
        #
        # everything that is not dependent on phi0:
        integrand *= np.exp(1j * kx * (
            r[0] * np.cos(phi0) +
            r[1] * np.sin(phi0))
        )

        # Calculate the integral for the position r
        integrand.sort()
        f[j] = np.sum(integrand)

        # Display how far we are
        if count is not None:
            with count.get_lock():
                count.value += 1

    return f.reshape((N, N))



import numpy as np


def compute_angle_weights_1d(angles: np.ndarray) -> np.ndarray:
    """Compute angular weights for tomographic reconstruction

    This method aims to address the issue of unevenly distributed angles
    in tomographic reconstruction. For algorithms, such as filtered
    backprojection, weighting each backprojection with a factor
    proportional to the angular distance to its neighbors dramatically
    improves the reconstructed image.

    Weights computation also takes into account these special cases:

    - Same angle present multiple times
    - Angular coverage larger than 180° (PI)

    Parameters
    ----------
    angles: 1d ndarray of length A
        Angles corresponding to the projections [rad]]

    Returns
    -------
    weights: 1d ndarray of length A
        Weights for each angle; the mean of `weights` is one.

    Notes
    -----
    If one angle is passed multiple times `N` (e.g. `N=2`, 0° and 180°
    via `np.linspace(0, 180, 10, endpoint=True)`), then this angle will
    have a weight smaller than the other angles by a factor of `1/N`.
    """
    # copy and modulo np.pi
    # This is an array with values in [0, np.pi)
    angles = (angles.flatten() - np.min(angles)) % np.pi

    # If we have duplicate entries, we need to take them into account
    unq_angles, unq_reverse, unq_counts = np.unique(
        angles, return_inverse=True, return_counts=True)

    # sort the array
    srt_idx = np.argsort(unq_angles)
    srt_angles = unq_angles[srt_idx]
    srt_count = unq_counts[srt_idx]

    # compute weights for sorted angles
    da = (np.roll(srt_angles, -1) - np.roll(srt_angles, 1)) % np.pi
    sum_da = np.sum(da)

    # normalize with number of occurrences
    da /= srt_count
    srt_weights = da / sum_da * angles.size

    # Sort everything back where it belongs
    unq_weights = np.zeros_like(srt_weights)
    unq_weights[srt_idx] = srt_weights

    # Set the weights for each item in the original angles
    weights = np.zeros_like(angles)
    weights = unq_weights[unq_reverse]
    return weights



"""Volumetric data reconstruction benchmark

This simple example can be used to quantify the speed-up due to
multiprocessing on the hardware used. Note that
``multiprocessing.cpu_count`` does not return the number of
physical cores.
"""
from multiprocessing import cpu_count
import time

import numpy as np

import radontea as rt


A = 70    # number of angles
N = 128   # detector size x
M = 24    # detector size y (number of slices)

# generate random data
sino0 = np.random.random((A, N))     # for 2d example
sino = np.random.random((A, M, N))    # for 3d example
sino[:, 0, :] = sino0
angles = np.linspace(0, np.pi, A)     # for both

a = time.time()
data1 = rt.backproject_3d(sino, angles, ncpus=1)
print("time on 1 core:  {:.2f} s".format(time.time() - a))

a = time.time()
data2 = rt.backproject_3d(sino, angles, ncpus=cpu_count())
print("time on {} cores: {:.2f} s".format(cpu_count(), time.time() - a))

assert np.all(data1 == data2), "2D and 3D results don't match"



# flake8: noqa F401
from ._alg_bpj import backproject, backproject_3d
from ._alg_fmp import fourier_map, fourier_map_3d
from ._alg_int import integrate
from ._alg_art import art
from ._alg_sart import sart

from ._rdn_prl import radon_parallel
from ._threed import volume_recon

from ._version import version as __version__

from . import fan
from . import util


__author__ = "Paul Müller"
__license__ = "BSD (3-Clause)"



import multiprocessing as mp
import numpy as np
import time

from typing import Callable


def do_work(in_queue, out_list, count, max_count):
    while True:
        item = in_queue.get()
        # exit signal
        if item == "STOP":
            return

        func2d, kwargs, index = item
        result = func2d(count=count, max_count=max_count, **kwargs)

        out_list.append((index, result))
        time.sleep(.01)


def volume_recon(func2d: Callable, sinogram: np.ndarray = None,
                 angles: np.ndarray = None,
                 count=None, max_count=None, ncpus=None,
                 **kwargs) -> np.ndarray:
    """Slice-wise 3D inversion of the Radon transform

    Computes the slice-wise 3D inverse of the Radon transform using
    multiprocessing.


    Parameters
    ----------
    func2d: callable
        A method for the slice-wise reconstruction
        (e.g. :func:`backproject`).
    sinogram: ndarray, shape (A,M,N)
        Three-dimensional sinogram of line recordings. The axis `1`
        iterates through the `M` slices. The rotation takes place through
        axis `1`.
    angles: (A,) ndarray
        Angular positions of the `sinogram` in radians in the
        interval [0, PI).
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of this function. The
        value of `max_count.value` is set initially and the value
        of `count.value` is incremented until it reaches the end
        of the algorithm (`max_count.value`).
    **kwargs: dict
        Additional keyword arguments to `func2d`.

    Returns
    -------
    out: ndarray, shape (N,M,N)
        The reconstructed volume.

        .. versionchanged:: 0.4.0

            Output indexing now follows the ODTbrain convention. For the
            the old behavior, use ``out.transpose(1, 0, 2)``.
    """
    if sinogram.shape[0] != angles.shape[0]:
        msg = "First dimension of `sinogram` must match size of `angles`"
        raise ValueError(msg)

    if len(sinogram.shape) != 3:
        msg = "`sinogram` must have three dimensions."
        raise ValueError(msg)

    if ncpus is None:
        ncpus = mp.cpu_count()

    fkw = kwargs.copy()
    fkw["angles"] = angles

    manager = mp.Manager()
    results = manager.list()
    work = manager.Queue()

    # kick-off working processes
    pool = []
    max_counts = [mp.Value("I", 0, lock=True) for _ii in range(ncpus)]
    for ii in range(ncpus):
        p = mp.Process(target=do_work, args=(work, results,
                                             count, max_counts[ii]))
        p.start()
        pool.append(p)

    # add first slice job (we need to do this to get an estimate of max_count)
    fkwj = fkw.copy()
    fkwj["sinogram"] = sinogram[:, 0, :]
    work.put((func2d, fkwj, 0))

    # determine max_count for a single slice and set it globally
    if max_count is not None:
        for _ii in range(60):  # wait max 6s
            time.sleep(.01)
            initial_max_count = np.max([c.value for c in max_counts])
            if initial_max_count != 0:
                break
        with max_count.get_lock():
            max_count.value = sinogram.shape[1] * initial_max_count

    # add other slice jobs
    for jj in range(1, sinogram.shape[1]):
        fkwj = fkw.copy()
        fkwj["sinogram"] = sinogram[:, jj, :]
        work.put((func2d, fkwj, jj))

    # put stop signal at end of queue to let workers know they may stop
    for _kk in range(ncpus):
        work.put("STOP")

    for p in pool:
        p.join()

    sh = sinogram.shape
    out = np.zeros((sh[2], sh[1], sh[2]))
    for ii in range(len(results)):
        idx, res = results[ii]
        out[:, idx, :] = res

    return out



"""Comparison of parallel-beam reconstruction methods

This example illustrates the performance of the
different reconstruction techniques for a parallel-beam
geometry. The left column shows the reconstruction of
the original image and the right column shows the reconstruction
of the corresponding binary images. Note that the
SART process could be sped-up by computing an
initial guess with a non-iterative method and
setting it with the ``initial`` keyword argument.
"""
from matplotlib import pylab as plt
import numpy as np

import radontea
from radontea.logo import get_original

N = 55  # image size
A = 13  # number of sinogram angles
ITA = 10  # number of iterations a
ITB = 100  # number of iterations b

angles = np.linspace(0, np.pi, A)

im = get_original(N)
sino = radontea.radon_parallel(im, angles)
fbp = radontea.backproject(sino, angles)
fintp = radontea.fourier_map(sino, angles).real
sarta = radontea.sart(sino, angles, iterations=ITA)
sartb = radontea.sart(sino, angles, iterations=ITB)

im2 = (im >= (im.max() / 5)) * 255
sino2 = radontea.radon_parallel(im2, angles)
fbp2 = radontea.backproject(sino2, angles)
fintp2 = radontea.fourier_map(sino2, angles).real
sarta2 = radontea.sart(sino2, angles, iterations=ITA)
sartb2 = radontea.sart(sino2, angles, iterations=ITB)

plt.figure(figsize=(8, 22))
pltkw = {"vmin": -20,
         "vmax": 280}

plt.subplot(6, 2, 1, title="original image")
plt.imshow(im, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 2, title="binary image")
plt.imshow(im2, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 3, title="sinogram (from original)")
plt.imshow(sino)
plt.axis('off')

plt.subplot(6, 2, 4, title="sinogram (from binary)")
plt.imshow(sino2)
plt.axis('off')

plt.subplot(6, 2, 5, title="filtered backprojection")
plt.imshow(fbp, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 6, title="filtered backprojection")
plt.imshow(fbp2, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 7, title="Fourier interpolation")
plt.imshow(fintp, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 8, title="Fourier interpolation")
plt.imshow(fintp2, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 9, title="SART ({} iterations)".format(ITA))
plt.imshow(sarta, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 10, title="SART ({} iterations)".format(ITA))
plt.imshow(sarta2, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 11, title="SART ({} iterations)".format(ITB))
plt.imshow(sartb, **pltkw)
plt.axis('off')

plt.subplot(6, 2, 12, title="SART ({} iterations)".format(ITB))
plt.imshow(sartb2, **pltkw)
plt.axis('off')


plt.tight_layout()
plt.show()



import multiprocessing as mp

import numpy as np

import radontea


def test_simple():
    """Simple test for backproject"""
    count = mp.Value("i")
    max_count = mp.Value("i")
    sino = np.random.random((10, 30))
    radontea.backproject(sinogram=sino,
                         angles=np.arange(len(sino)),
                         count=count,
                         max_count=max_count)
    assert count.value == 11
    assert count.value == max_count.value


def test_max_count_all_parralel_2d():
    """Test that count always reaches max_count"""
    sino = np.random.random((10, 31))
    angles = np.arange(len(sino))
    for name in ["backproject",
                 "fourier_map",
                 "integrate",
                 "art",
                 "sart",
                 ]:

        method = getattr(radontea, name)
        count = mp.Value("i")
        max_count = mp.Value("i")
        method(sinogram=sino,
               angles=angles,
               count=count,
               max_count=max_count)
        assert max_count.value == count.value


def test_bpj_3d_convenience():
    """test that max_count is incremented correctly in 3D methods"""
    sino = np.random.random((10, 4, 31))
    angles = np.arange(len(sino))
    count = mp.Value("i")
    max_count = mp.Value("i")
    radontea.backproject_3d(sinogram=sino,
                            angles=angles,
                            count=count,
                            max_count=max_count)
    assert max_count.value == 44  # and not 11
    assert max_count.value == count.value


def test_fmp_3d_convenience():
    """test that max_count is incremented correctly in 3D methods"""
    sino = np.random.random((10, 4, 31))
    angles = np.arange(len(sino))
    count = mp.Value("i")
    max_count = mp.Value("i")
    radontea.fourier_map_3d(sinogram=sino,
                            angles=angles,
                            count=count,
                            max_count=max_count)
    assert max_count.value == 16  # and not 4
    assert max_count.value == count.value


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Progress monitoring with progression

The progress of the reconstruction algorithms in radontea
can be tracked with other packages such as
`progression <https://cimatosa.github.io/progression/>`_.
"""
from multiprocessing import cpu_count

import numpy as np
import progression as pr

import radontea as rt


A = 55    # number of angles
N = 128   # detector size x
M = 24    # detector size y (number of slices)

# generate random data
sino0 = np.random.random((A, N))
sino = np.random.random((A, M, N))
sino[:, 0, :] = sino0
angles = np.linspace(0, np.pi, A)

count = pr.UnsignedIntValue()
max_count = pr.UnsignedIntValue()


with pr.ProgressBar(count=count,
                    max_count=max_count,
                    interval=0.3
                    ) as pb:
    pb.start()
    rt.volume_recon(func2d=rt.sart, sinogram=sino, angles=angles,
                    ncpus=cpu_count(), count=count, max_count=max_count)



import numpy as np
import scipy.ndimage


def radon_parallel(arr: np.ndarray, angles: np.ndarray,
                   count=None, max_count=None,) -> np.ndarray:
    """Compute the Radon transform (sinogram) of a circular image.

    The :mod:`scipy` Radon transform performs this operation on the
    entire image, whereas this implementation requires an input
    image that has gray-scale values of 0 outside of a circle
    with diameter equal to the image size.


    Parameters
    ----------
    arr: ndarray, shape (N,N)
        the input image.
    angles: ndarray, length A
        angles or projections in radians
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.


    Returns
    -------
    outarr: ndarray of floats, shape (A, N)
        Sinogram of the input image. The i'th row contains the
        projection data of the i'th angle.


    See Also
    --------
    scipy.ndimage.interpolation.rotate :
        The interpolator used to rotate the image.
    """
    # This function also works with single angles
    angles = np.atleast_1d(angles)
    A = angles.shape[0]
    N = len(arr)
    # The radon function from skimage.transform does not allow
    # to reshape the image (one could cut it of course).
    # outarray: x-axis: projection
    #           y-axis:
    outarr = np.zeros((A, N))

    # progress monitoring
    if max_count is not None:
        with max_count.get_lock():
            max_count.value += A + 1
    if count is not None:
        with count.get_lock():
            count.value += 1

    for i in np.arange(A):
        rotated = scipy.ndimage.rotate(arr,
                                       angles[i] / np.pi * 180,
                                       order=3,
                                       reshape=False,
                                       mode="constant",
                                       cval=0)
        # sum along the axis.
        outarr[i] = rotated.sum(axis=0)
        if count is not None:
            with count.get_lock():
                count.value += 1
    return outarr



import warnings

import numpy as np
import scipy.interpolate as intp

from . import _threed


def fourier_map(sinogram: np.ndarray, angles: np.ndarray,
                intp_method: str = "cubic",
                count=None, max_count=None) -> np.ndarray:
    """2D Fourier mapping with the Fourier slice theorem

    Computes the inverse of the Radon transform using Fourier
    interpolation.
    Warning: This is the naive reconstruction that assumes that
    the image is rotated through the upper left pixel nearest to
    the actual center of the image. We do not have this problem for
    odd images, only for even images.


    Parameters
    ----------
    sinogram: (A,N) ndarray
        Two-dimensional sinogram of line recordings.
    angles: (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    intp_method: {'cubic', 'nearest', 'linear'}, optional
        Method of interpolation. For more information see
        `scipy.interpolate.griddata`. One of

          - "nearest": instead of interpolating, use the points closest
            to the input data
          - "linear": bilinear interpolation between data points
          - "cubic": interpolate using a two-dimensional poolynimial
            surface

    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.


    Returns
    -------
    out: ndarray
        The reconstructed image.


    See Also
    --------
    scipy.interpolate.griddata: interpolation method used
    """
    if max_count is not None:
        with max_count.get_lock():
            max_count.value += 4

    if len(sinogram[0]) % 2 == 0:
        warnings.warn("Fourier interpolation with slices that have" +
                      " even dimensions leads to image distortions!")
    # projections
    p_x = sinogram

    # Fourier transform of the projections
    # The sinogram data is shifted in Fourier space
    P_fx = np.fft.fft(np.fft.ifftshift(p_x, axes=-1), axis=-1)

    if count is not None:
        with count.get_lock():
            count.value += 1

    # This paragraph could be useful for future version if the reconstruction
    # grid is to be changed. They directly affect *P_fx*.
    # if False:
    #    # Resize everyting
    #    factor = 10
    #    P_fx = np.zeros((len(sinogram), len(sinogram[0])*factor),
    #                    dtype=np.complex128)
    #    newp = np.zeros((len(sinogram), len(sinogram[0])*factor),
    #                    dtype=np.complex128)
    #    for i in range(len(sinogram)):
    #        x = np.linspace(0, len(sinogram[0]), len(sinogram[0]))
    #        dint = intp.interp1d(x, sinogram[i])
    #        xn = np.linspace(0, len(sinogram[0]), len(sinogram[0])*factor)
    #        datan = dint(xn)
    #        datfft = np.fft.fft(datan)
    #        newp[i] = datan
    #        # shift datfft
    #        P_fx[i] = np.roll(1*datfft,0)
    #    p_x = newp
    # if False:
    #    # Resize the input image
    #    P_fx = np.zeros(p_x.shape, dtype=np.complex128)
    #    for i in range(len(sinogram)):
    #        factor = 2
    #        x = np.linspace(0, len(sinogram[0]), len(sinogram[0]))
    #        dint = intp.interp1d(x, sinogram[i])#, kind="nearest")
    #        xn = np.linspace(0, len(sinogram[0]), len(sinogram[0])*factor)
    #        datan = dint(xn)
    #        datfft = np.fft.fft(datan)
    #        fftint = intp.interp1d(xn, datfft)
    #        start = (len(xn) - len(x))/2
    #        end = (len(xn) + len(x))/2
    #        fidata = fftint(x)
    #        datfft = np.fft.fftshift(datfft)
    #        datfft = datfft[start:end]
    #        datfft = np.fft.ifftshift(datfft)
    #        P_fx[i] = 1*datfft

    # angles need to be normalized to 2pi
    # if angles star with 0, then the image is falsly rotated
    # unfortunately we still have ashift in the data.
    ang = (angles.reshape(-1, 1))

    # compute frequency coordinates fx
    fx = np.fft.fftfreq(len(p_x[0]))

    fx = fx.reshape(1, -1)
    # fy is zero
    fxl = (fx) * np.cos(ang)
    fyl = (fx) * np.sin(ang)
    # now fxl, fyl, and P_fx have same shape

    # DEBUG: plot coordinates of positions of projections in fourier domain
    #   from matplotlib import pylab as plt
    #   plt.figure()
    #   for i in range(len(fxl)):
    #      plt.plot(fxl[i],fyl[i],"x")
    #   plt.axes().set_aspect('equal')
    #   plt.show()

    # flatten everything for interpolation
    Xf = fxl.flatten()
    Yf = fyl.flatten()
    Zf = P_fx.flatten()
    # rintp defines the interpolation grid
    rintp = np.fft.fftshift(fx.reshape(-1))

    if count is not None:
        with count.get_lock():
            count.value += 1

    # The code block yields the same result as griddata (below)
    # interpolation coordinates
    #   Rf = np.zeros((len(Xf),2))
    #   Rf[:,0] = 1*Xf
    #   Rf[:,1] = 1*Yf
    # reconstruction coordinates
    #   Nintp = len(rintp)
    #   Xn, Yn = np.meshgrid(rintp,rintp)
    #   Rn = np.zeros((Nintp**2, 2))
    #   Rn[:,0] = Xn.flatten()
    #   Rn[:,1] = Yn.flatten()
    #
    #    if intp_method.lower() == "bilinear":
    #       Fr = intp.LinearNDInterpolator(Rf,Zf.real)
    #       Fi = intp.LinearNDInterpolator(Rf,Zf.imag)
    #    elif intp_method.lower() == "nearest":
    #       Fr = intp.NearestNDInterpolator(Rf,Zf.real)
    #       Fi = intp.NearestNDInterpolator(Rf,Zf.imag)
    #    else:
    #       raise NotImplementedError("Unknown interpolation type: {}".format(
    #                                  intp_method.lower()))
    #   Fcomp = (Fr(Rn) + 1j*Fi(Rn)).reshape(Nintp,Nintp)

    # The actual interpolation
    Fcomp = intp.griddata((Xf, Yf), Zf, (rintp[None, :], rintp[:, None]),
                          method=intp_method)

    if count is not None:
        with count.get_lock():
            count.value += 1

    # remove nans
    Fcomp[np.where(np.isnan(Fcomp))] = 0

    f = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(Fcomp)))

    if count is not None:
        with count.get_lock():
            count.value += 1

    return f.real


def fourier_map_3d(sinogram: np.ndarray, angles: np.ndarray,
                   intp_method: str = "cubic",
                   count=None, max_count=None, ncpus=None) -> np.ndarray:
    """Convenience wrapper for 3D Fourier mapping reconstruction

    See :func:`fourier_map` for parameter definitions. The additional
    parameter `ncpus` sets the number of CPUs used.

    Returns
    -------
    out: ndarray, shape (N,M,N)
        The reconstructed volume.

        .. versionchanged:: 0.4.0

            Output indexing now follows the ODTbrain convention. For the
            the old behavior, use ``out.transpose(1, 0, 2)``.
    """
    return _threed.volume_recon(func2d=fourier_map,
                                sinogram=sinogram,
                                angles=angles,
                                intp_method=intp_method,
                                count=count,
                                max_count=max_count,
                                ncpus=ncpus)



import numpy as np


def art(sinogram: np.ndarray, angles: np.ndarray, initial: np.ndarray = None,
        iterations: int = 1, count=None, max_count=None) -> np.ndarray:
    """Algebraic Reconstruction Technique

    The Algebraic Reconstruction Technique (ART) iteratively
    computes the inverse of the Radon transform in two dimensions.
    The reconstruction technique uses *rays* of the diameter of
    one pixel to iteratively solve the system of linear equations
    that describe the projection process. The binary weighting
    factors are

     - 1, if the center of the a pixel is within the *ray*
     - 0, else


    Parameters
    ----------
    sinogram: ndarrayy, shape (A,N)
        Two-dimensional sinogram of line recordings.
    angles: ndarray, length A
        Angular positions of the `sinogram` in radians. The angles
        at which the sinogram slices were recorded do not have to be
        distributed equidistantly as in :func:`backproject`.
        The angles are internaly converted to modulo PI.
    initial: ndarray, shape (N,N), optional
        The initial guess for the solution.
    iterations: int
        Number of iterations to perform.
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.


    Notes
    -----
    For theoretical backround, see
    Kak, A. C., & Slaney, M.. *Principles of Computerized
    Tomographic Imaging*, SIAM, (2001)

    Sec. 7.2:
    *"ART reconstrutions usually suffer from salt and pepper noise,
    which is caused by the inconsitencies introuced in the set of
    equations by the approximations commonly used for*
    :math:`w_{ik}` *'s."*
    """
    # make sure `iterations` is an integer
    iterations = int(iterations)
    N = sinogram.shape[1]
    A = angles.shape[0]

    if max_count is not None:
        with max_count.get_lock():
            max_count.value += A * iterations + 1
    # Meshgrid for weigths
    center = N / 2.0
    x = np.arange(N) - center + .5
    X, Y = np.meshgrid(x, x)
    # initiate array
    if initial is None:
        f = np.zeros((N, N), dtype=np.dtype(float))
    else:
        f = 1 * initial.transpose()[::-1]
    # Make sure all angles are in [0,PI)
    for i in np.arange(A):
        if angles[i] > np.pi:
            offset = np.floor(angles[i] / np.pi)
            angles[i] -= offset * np.pi
        elif angles[i] < 0:
            offset = np.floor(np.abs((angles[i] / np.pi))) + 1
            angles[i] += offset * np.pi
        if angles[i] == np.pi:
            angles[i] = 0

    # These lambda functions return the two x- and y- values of two
    # points projected onto a line along the x- and y-axis, having the
    # angle angle_k in radians.
    # Sourcing this out to here makes things a little faster.
    def GetLambdaLines(angle_k):
        if angle_k == 0:
            # Divide by zero error for tan
            # We have horizontal lines (parallel to x)
            def line1(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k): return (
                x2 - .5, ypi1 * np.ones(y2.shape))

            def line2(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k): return (
                x2 + .5, ypi2 * np.ones(y2.shape))
        elif angle_k == np.pi / 2:
            # We have vertical lines (parallel to y)
            def line1(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k): return (
                xpi1 * np.ones(x2.shape), y2 + .5)

            def line2(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k): return (
                xpi2 * np.ones(x2.shape), y2 - .5)
        elif angle_k < np.pi / 2:
            # CASE 1
            # Compute any other positions on the lines from the given things
            def line1(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k):
                return ((y2 - ypi1) / np.tan(angle_k) + xpi1,
                        (x2 - xpi1) * np.tan(angle_k) + ypi1)
            # def line2(x2,y2):
            #    y = (x2-xpi2)/np.tan(angle_k) - ypi2
            #    x = (y2-ypi2)*np.tan(angle_k) - xpi2
            #    return x,y

            def line2(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k):
                return ((y2 - ypi2) / np.tan(angle_k) + xpi2,
                        (x2 - xpi2) * np.tan(angle_k) + ypi2)
        else:
            # CASE 2: Switch x-output - only for speed.
            # Not very obvious - possibly a hack.
            # Compute any other positions on the lines from the given things
            def line1(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k):
                return ((y2 - ypi2) / np.tan(angle_k) + xpi2,
                        (x2 - xpi1) * np.tan(angle_k) + ypi1)
            # def line2(x2,y2):
            #    y = (x2-xpi2)/np.tan(angle_k) - ypi2
            #    x = (y2-ypi2)*np.tan(angle_k) - xpi2
            #    return x,y

            def line2(x2, y2, xpi1, ypi1, xpi2, ypi2, angle_k):
                return ((y2 - ypi1) / np.tan(angle_k) + xpi1,
                        (x2 - xpi2) * np.tan(angle_k) + ypi2)
        return line1, line2
    # Sort angles?
    # This could increase the speed of convergence.

    # f[j] is consistent with Kak, Slaney
    f = f.flatten()

    if count is not None:
        with count.get_lock():
            count.value += 1

    for iteration in np.arange(iterations):  # @UnusedVariable
        #
        # i iterates the rays
        # j iterates the flattened outarr
        # k iterates the angles
        #
        for k in np.arange(A):
            # From now on we work in radians
            angle_k = angles[k]
            # p[i] is consistent with Kak, Slaney
            p = sinogram[k]
            # # We do not store the binary weights for each ray:
            # # For large images this thing could get big:
            # w = np.zeros((len(p),N*N), dtype=bool)
            # # w[i][j]
            line1, line2 = GetLambdaLines(angle_k)
            # CASES for all 2 quadrants. Angles are modulo PI
            # This case stuff is dependent on the angle. We enumerate
            # pr in negative mathematical angular direction
            #
            # This is the position on the projection, centered around 0:
            pr = np.arange(len(p)) - center + .5  # radial distance,
            #
            if angle_k <= np.pi / 2:
                # 0 to PI/2
                # case == 1
                # position of each p[i] in the centered *outarr*.
                # a position of line1 in X
                x_p1 = (pr - .5) * np.sin(angle_k)
                # a position of line1 in Y
                y_p1 = -(pr - .5) * np.cos(angle_k)
                # a position of line2 in X
                x_p2 = (pr + .5) * np.sin(angle_k)
                # a position of line2 in Y
                y_p2 = -(pr + .5) * np.cos(angle_k)
            else:
                # PI/2 to PI
                # case == 2
                # position of each p[i] in the centered *outarr*.
                # a position of line1 in X
                x_p1 = (pr + .5) * np.sin(angle_k)
                # a position of line1 in Y
                y_p1 = -(pr + .5) * np.cos(angle_k)
                # a position of line2 in X
                x_p2 = (pr - .5) * np.sin(angle_k)
                # a position of line2 in Y
                y_p2 = -(pr - .5) * np.cos(angle_k)
            for i in np.arange(len(p)):
                # If the angle is zero, then we are looking at the
                # projections onto the right side. The indices are
                # enumerated from bottom to top.
                w_i = np.zeros((N, N), dtype=bool)
                # # where does the ray cut w_i?
                # xpi1 = x_p1[i]
                # ypi1 = y_p1[i]
                # xpi2 = x_p2[i]
                # ypi2 = y_p2[i]
                # Check if a value is between the two lines (within ray)
                # For each value of X and Y, compute what the values of the
                # line would be.
                xl1, yl1 = line1(
                    X, Y, x_p1[i], y_p1[i], x_p2[i], y_p2[i], angle_k)
                xl2, yl2 = line2(
                    X, Y, x_p1[i], y_p1[i], x_p2[i], y_p2[i], angle_k)
                #
                AND = np.logical_and
                # Values lie between the two lines if the following is True:
                # Furthermore we restrict ourselves to a disk.
                w_i = AND(X**2 + Y**2 < center**2,
                          AND(AND(xl1 < X, X < xl2),
                              AND(yl1 > Y, Y > yl2))).flatten()
                #
                # i iterates the rays
                # j iterates the flattened outarr
                # k iterates the angles
                # In each iteration of the angles the image is changed
                # if np.sum(w_i) != 0:
                #    f += ( p[i] - np.sum(f*w_i) )/np.sum(w_i) * w_i
                # This is faster because we don't a the zeros.
                valid = np.where(w_i)
                f[valid] += (p[i] - np.sum(f[valid])) / np.sum(w_i)
            if count is not None:
                with count.get_lock():
                    count.value += 1
    # By slicing in-place [::-1] we get rid of the inversion of the
    # image along the y-axis.

    return f.reshape((N, N))[::-1].transpose()



import numpy as np

import radontea


def test_radon_simple():
    x = np.zeros((10, 10))
    x[2:5, 3:6] = 1
    sino = radontea.radon_parallel(x, angles=[0, np.pi/2, np.pi, np.pi*3/2])
    assert np.allclose(sino[0], [0, 0, 0, 3, 3, 3, 0, 0, 0, 0])
    assert np.allclose(sino[1], [0, 0, 3, 3, 3, 0, 0, 0, 0, 0])
    assert np.allclose(sino[2], [0, 0, 0, 0, 3, 3, 3, 0, 0, 0])
    assert np.allclose(sino[3], [0, 0, 0, 0, 0, 3, 3, 3, 0, 0])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

import radontea.util as util


def test_even():
    angles = np.linspace(0, np.pi, 18, endpoint=False)
    res = util.compute_angle_weights_1d(angles)
    assert np.allclose(res, 1, rtol=0, atol=1e-14)


def test_same_end_value():
    angles = np.linspace(0, np.pi, 18, endpoint=True)
    res = util.compute_angle_weights_1d(angles)
    assert np.allclose(np.mean(res), 1, rtol=0, atol=1e-14)
    assert np.allclose(res[1:-1], 18 / 17, rtol=0, atol=1e-14)
    assert np.allclose(res[0], 18 / 17 / 2, rtol=0, atol=1e-14)
    assert np.allclose(res[-1], 18 / 17 / 2, rtol=0, atol=1e-14)


def test_multiple_identical_angles():
    angles_0 = np.linspace(0, np.pi, 14, endpoint=False)
    angles_1 = np.roll(angles_0, -3)
    angles_2 = np.concatenate((np.ones(4)*angles_1[0], angles_1))
    angles = np.roll(angles_2, 3)
    res = util.compute_angle_weights_1d(angles)
    assert np.allclose(np.mean(res), 1, rtol=0, atol=1e-14)
    assert np.allclose(res[:3], 18 / 14, rtol=0, atol=1e-14)
    assert np.allclose(res[3+4+1:], 18 / 14, rtol=0, atol=1e-14)
    assert np.allclose(res[3:8], 18 / 14 / 5, rtol=0, atol=1e-14)



"""Creates a sinogram for testing purposes"""
import numpy as np


def create_test_sino(A, N):
    """
    Creates test sinogram.
    """
    resar = np.zeros((A, N), dtype=float)
    angles = np.linspace(0, np.pi, A, endpoint=False)
    x = np.linspace(-N/2, N/2, N, endpoint=True)
    dev = np.sqrt(N/2)
    off = N/7
    for ii in range(A):
        # Gaussian distribution sinogram
        x0 = np.cos(angles[ii])*off
        y = np.exp(-(x-x0)**2/dev**2)
        resar[ii] = y

    # some normalization, so that max of Radon-inversion
    # is about 1.
    resar = 2 * dev * resar / resar.max()
    return resar, angles



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import exists, dirname, realpath
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'Collection of algorithms to compute the inverse Radon transform'
name = 'radontea'
year = "2014"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version  # noqa: E402


setup(
    name=name,
    author=author,
    author_email="dev@craban.de",
    version=version,
    license="BSD (3 clause)",
    url='https://github.com/RI-imaging/radontea',
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["numpy>=1.5.1",
                      "scipy>=1.4.0",  # Updated QHull in griddata
                      ],
    python_requires='>=3.5, <4',
    keywords=["tomography", "ct", "radon", "computerized tomography",
              "optical projection tomography"],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )



import pathlib

import numpy as np

import radontea
import sinogram


def test_2d_int():
    sino, angles = sinogram.create_test_sino(A=20, N=30)
    r = radontea.integrate(sino, angles)

    # np.savetxt('outfile.txt', np.array(r).flatten().view(float), fmt="%.8f")
    reffile = pathlib.Path(__file__).parent / "data" / "2d_int.txt"
    ref = np.loadtxt(str(reffile))
    assert np.allclose(np.array(r).flatten().view(float), ref)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np
import scipy.interpolate as intp

from . import _threed, util


def backproject(sinogram: np.ndarray, angles: np.ndarray,
                filtering: str = "ramp", weight_angles: bool = True,
                padding: bool = True, padval: float = 0, count=None,
                max_count=None, verbose: int = 0) -> np.ndarray:
    r"""2D backprojection algorithm

    Computes the inverse of the Radon transform using filtered
    backprojection.


    Parameters
    ----------
    sinogram: ndarray, shape (A,N)
        Two-dimensional sinogram of line recordings.
    angles: (A,) ndarray
        Angular positions of the `sinogram` in radians.
    filtering: {'ramp', 'shepp-logan', 'cosine', 'hamming', \
                 'hann'}, optional
        Specifies the Fourier filter. Either of

          - "ramp": mathematically correct reconstruction
          - "shepp-logan"
          - "cosine"
          - "hamming"
          - "hann"

    weight_angles: bool
        If ``True``, weights each backpropagated projection with a factor
        proportional to the angular distance between the neighboring
        projections.

        .. math::
            \Delta \phi_0 \longmapsto \Delta \phi_j =
            \frac{\phi_{j+1} - \phi_{j-1}}{2}

        .. versionadded:: 0.1.9

    padding: bool, optional
        Pad the input data to the second next power of 2 before
        Fourier transforming. This reduces artifacts and speeds up
        the process for input image sizes that are not powers of 2.
    padval: float
        The value used for padding.
        If `padval` is `None`, then the edge values are used for
        padding (see documentation of `numpy.pad`).
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.
    verbose: int
        Increment to increase verbosity.

    Returns
    -------
    out: ndarray
        The reconstructed image.


    Examples
    --------
    >>> import numpy as np
    >>> from radontea import backproject
    >>> N = 5
    >>> A = 7
    >>> x = np.linspace(-N/2, N/2, N)
    >>> projection = np.exp(-x**2)
    >>> sinogram = np.tile(projection, A).reshape(A,N)
    >>> angles = np.linspace(0, np.pi, A, endpoint=False)
    >>> backproject(sinogram, angles)
    array([[ 0.03813283, -0.01972347, -0.02221885,  0.03822303,  0.01903376],
           [ 0.04033526, -0.01591647,  0.02262173,  0.04203002,  0.02123619],
           [ 0.02658797,  0.11375576,  0.41525594,  0.17170226,  0.0074889 ],
           [ 0.04008672,  0.11139209,  0.27650193,  0.16933859,  0.02098764],
           [ 0.02140445,  0.0334597 ,  0.07691547,  0.0914062 ,  0.00230537]])
    """
    # Perform weighting
    if weight_angles:
        weights = util.compute_angle_weights_1d(angles).reshape(-1, 1)
        sinogram = sinogram * weights
    else:
        sinogram = sinogram

    ln = sinogram.shape[1]
    la = angles.shape[0]
    assert sinogram.shape[0] == la

    if max_count is not None:
        with max_count.get_lock():
            max_count.value += la + 1

    # We perform padding before performing the Fourier transform.
    # This gets rid of artifacts due to false periodicity and also
    # speeds up Fourier transforms of the input image size is not
    # a power of 2.

    if padding:
        order = max(64., 2**np.ceil(np.log(ln * 2.1) / np.log(2)))
        pad = order - ln
    else:
        pad = 0
        order = ln

    padl = int(np.ceil(pad / 2))
    padr = int(pad - padl)

    if padval is None:
        if verbose > 0:
            print("......Padding with edge values.")
        sino = np.pad(sinogram, ((0, 0), (padl, padr)),
                      mode="edge")
        sino = np.roll(sino, -padl, 1)
    else:
        if verbose > 0:
            print("......Verifying padding value: {}".format(padval))
        sino = np.pad(sinogram, ((0, 0), (padl, padr)),
                      mode="linear_ramp",
                      end_values=(padval,))
        sino = np.roll(sino, -padl, 1)

    # These artifacts are for example bad contrast.
    kx = 2 * np.pi * np.abs(np.fft.fftfreq(int(order)))
    # Ask for the filter. Do not include zero (first element).
    if filtering == "ramp":
        pass
    elif filtering == "shepp-logan":
        kx[1:] = kx[1:] * np.sin(kx[1:]) / (kx[1:])
    elif filtering == "cosine":
        kx[1:] = kx[1:] * np.cos(kx[1:])
    elif filtering == "hamming":
        kx[1:] = kx[1:] * (0.54 + 0.46 * np.cos(kx[1:]))
    elif filtering == "hann":
        kx[1:] = kx[1:] * (1 + np.cos(kx[1:])) / 2
    elif filtering is None:
        kx[1:] = 2 * np.pi
    else:
        raise ValueError("Unknown filter: %s" % filter)

    if count is not None:
        with count.get_lock():
            count.value += 1
    # Resize f so we can multiply it with the sinogram.
    kx = kx.reshape(1, -1)
    projection = np.fft.fft(sino, axis=-1) * kx
    sino_filtered = np.real(np.fft.ifft(projection, axis=-1))
    # Resize filtered sinogram back to original size
    sino = sino_filtered[:, :ln]
    # Perform backprojection
    x = np.linspace(-ln / 2, ln / 2, ln, endpoint=False) + .5
    outarr = np.zeros((ln, ln), dtype=np.float64)
    # Meshgrid for output array
    xv, yv = np.meshgrid(x, x)

    for ii in np.arange(len(angles)):
        projinterp = intp.interp1d(x, sino[ii], fill_value=0.0,
                                   copy=False, bounds_error=False)
        # Call proj_interp with cos and sin of corresponding coordinate
        # and add it to the outarr.
        phi = angles[ii]
        # Smear projection onto 2d volume
        xp = xv * np.cos(phi) + yv * np.sin(phi)
        outarr += projinterp(xp)
        # Here a much slower version with the same result:
        # for j in x:
        #    for k in x:
        #        # shift points to match origin of coordinate system
        #        xv = j - ln/2
        #        yv = k - ln/2
        #        # perform a simple coordinate transform
        #        #
        #        xp = xv*np.cos(phi) + yv*np.sin(phi)
        #        # yp = -xv*np.sin(phi) + yv*np.cos(phi)
        #        projval = projinterp(xp)
        #        outarr[j][k] += projval
        if count is not None:
            with count.get_lock():
                count.value += 1

    # Normalize output (we assume that the projections are equidistant)
    # We measure angles in radians
    dphi = np.pi / len(angles)
    # The factor of 2π comes from the choice of the unitary angular
    # frequency Fourier transform.
    outarr *= dphi / (2 * np.pi)

    return outarr


def backproject_3d(sinogram: np.ndarray, angles: np.ndarray,
                   filtering: str = "ramp", weight_angles: bool = True,
                   padding: bool = True, padval: float = 0, count=None,
                   max_count=None, ncpus=None) -> np.ndarray:
    """Convenience wrapper for 3D backprojection reconstruction

    See :func:`backproject` for parameter definitions. The additional
    parameter `ncpus` sets the number of CPUs used.

    Returns
    -------
    out: ndarray, shape (N,M,N)
        The reconstructed volume.

        .. versionchanged:: 0.4.0

            Output indexing now follows the ODTbrain convention. For the
            the old behavior, use ``out.transpose(1, 0, 2)``.
    """
    return _threed.volume_recon(func2d=backproject,
                                sinogram=sinogram,
                                filtering=filtering,
                                weight_angles=weight_angles,
                                padding=padding,
                                padval=padval,
                                angles=angles,
                                count=count,
                                max_count=max_count,
                                ncpus=ncpus)



"""Functions that are used to create the logo of radontea."""
import numpy as np

from radontea import radon_parallel


def logo(x, y, N) -> np.ndarray:
    """Vector representation of radontea logo.

    Parameters
    ----------
    x,y: ndarray or float
        coordinates where to calculate the logo.
    N: float
        total size of the image (NxN)


    Returns
    -------
    v: ndarray
        inverted values of the logo at the coordinates (x,y)
        normalized to one.
    """
    z1 = np.exp(-((x - N / 8)**2 + (y + N / 14)**2) / (N / 8)**2)
    z2 = np.exp(-((x + N / 10)**2 + (y - N / 10)**2) / (N / 8)**2)

    z3 = np.exp(-((x - N / 8)**2 + (y + N / 10)**2) / (N / 12)**2)
    z4 = np.exp(-((x + N / 10)**2 + (y - N / 14)**2) / (N / 12)**2)

    z5 = np.exp(-((x)**2 + (y)**2) / (N / 4)**2)
    z = (z1 + z2 - (z3 + z4))
    z = (z - z.min()) * z5
    z /= z.max()
    z[np.where(x**2 + y**2 >= (N / 2)**2)] = 0

    return z


def get_original(N: int = 64) -> np.ndarray:
    """radontea logo base image"""
    x = np.linspace(-N / 2, N / 2, N, endpoint=False)
    X = x.reshape(1, -1)
    Y = x.reshape(-1, 1)
    z = logo(X, Y, N)

    return np.array((z) * 255, dtype=np.uint16)


def get_logo(N: int = 64):
    """Return the radontea logo as a 2D NxN array

    This function discretizes the vector image representation of the
    radonteach logo.
    """
    a = (get_original(N=N))
    angles = np.linspace(0, np.pi * 1.6, N)
    sinogram = radon_parallel(a, angles)
    sinogram = (1 - sinogram / sinogram.max()) * 255
    return sinogram.transpose()


def main():
    # Show the logo
    from matplotlib import pylab as plt
    logo = get_logo(N=128)
    plt.imshow(logo, cmap="gray")
    plt.axis("off")
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main()



import numpy as np
import scipy.interpolate

from ._rdn_fan import get_fan_coords
from ._rdn_fan import get_det_coords, radon_fan  # noqa F401

from typing import Callable


def fan_rec(linogram: np.ndarray, lds: float, method: Callable, stepsize=1,
            det_spacing: float = 1, numang: int = None, count=None,
            max_count=None, **kwargs):
    """2D synthetic aperture reconstruction

    Computes the inverse of the fan-beam Radon transform using
    interpolation of the linogram and one of the inverse algorithms
    for tomography with the Fourier slice theorem.

    Parameters
    ----------
    linogram: 2d ndarray of shape (D, A)
        Input linogram from the synthetic aperture measurement.
    lds: float
        Distance in pixels between source and detector.
    method: callable
        Reconstruction method, e.g. `radontea.backproject`.
    numang: int
        Number of angles to be used for the sinogram. A higher number
        increases quality, but interpolation takes longer. By default
        numang = linogram.shape[1].
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.
    **kwargs: dict
        Keyword arguments for `method`.
    """
    sino, angles = lino2sino(linogram, lds, numang=numang, retang=True,
                             stepsize=stepsize, det_spacing=det_spacing,
                             count=count, max_count=max_count)
    return method(sino, angles, **kwargs)


def lino2sino(linogram: np.ndarray, lds: float, stepsize: float = 1,
              det_spacing: float = 1, numang: int = None,
              retang: bool = False,
              count=None, max_count=None) -> np.ndarray:
    """Convert linogram to sinogram for an equispaced detector.

    Parameters
    ----------
    linogram: real 2d ndarray of shape (D, A*)
        Linogram from synthetic aperture measurements.
    lds: float
        Distance from point source to detector in au.
    stepsize: float
        Translational increment of object in au (stepsize in D).
    det_spacing: float
        Distance between detector positions in au.
    numang: int
        Number of equispaced angles, defaults to linogram.shape[1]
    retang: bool
        Return the corresponding angles for the sinogram.
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.

    Returns
    -------
    sinogram: 2d ndarray of shape (D, A)
        The distortion-corrected sinogram.
        If retang is True, then the equispaced angles are returned as
        well.

    Notes
    -----
    This function can be used to convert a linogram obtained with
    fan-beam tomography to a sinogram, which then can be reconstructed
    with the backprojection or fourier mapping algorithms.
    """
    if not np.isreal(linogram):
        raise ValueError("`linogram` must be a real-valued array!")

    (D, det_size) = linogram.shape

    if numang is None:
        A = det_size  # (detector size determines # of angles)
    else:
        A = numang

    if max_count is not None:
        with max_count.get_lock():
            max_count.value += D + A

    # equispaced angles and corresponding lateral detector positions.
    angles, xang = get_fan_coords(det_size, det_spacing, lds, A)

    uorig = linogram
    lino = np.zeros((D, A))

    # go into the coordinate system of the original data
    xk = np.linspace(xang[0], xang[-1], det_size, endpoint=True)

    # begin equispacing
    for i in range(D):
        lino[i] = scipy.interpolate.spline(xk, uorig[i], xang)

        if count is not None:
            with count.get_lock():
                count.value += 1

    # begin angular stretching
    for i in range(A):
        # parametrization of the time axis (spatial information)
        xk = np.linspace(-D * stepsize / 2 + .5,
                         D * stepsize / 2 - .5, D, endpoint=True)
        alpha = angles[i]

        # Centering:
        # The object moves 1px per view in D.
        # We need to translate this lateral movement to a shift that
        # depends on the current angle.
        # What is the distance b/w the center of the object (centered at
        # lDS/2) to the axis alpha = 0?
        deltaD = np.tan(alpha) * lds / 2

        # Shearing:
        # At larger angles, the object seems bigger on the screen.
        xnew = xk / np.cos(alpha) + deltaD
        lino[:, i] = scipy.interpolate.spline(xk, lino[:, i], xnew)

        if count is not None:
            with count.get_lock():
                count.value += 1

    sino = np.transpose(lino)[:, ::-1]
    if retang:
        return sino, angles + np.pi / 2
    else:
        return sino
